<?php

namespace Drupal\webform_handler_defer\Plugin\WebformHandler;

use Drupal\Core\Form\FormStateInterface;
use Drupal\user\Entity\User;
use Drupal\webform\Entity\WebformSubmission;
use Drupal\webform\Plugin\WebformHandlerBase;
use Drupal\webform\WebformInterface;
use Drupal\webform\WebformSubmissionInterface;

/**
 * Handler which defers processing of other handlers.
 *
 * @WebformHandler(
 *   id = "webform_handler_defer",
 *   label = @Translation("Webform Handler Defer"),
 *   category = @Translation("Custom"),
 *   description = @Translation("Defers specified handlers until cron execution."),
 *   cardinality = \Drupal\webform\Plugin\WebformHandlerInterface::CARDINALITY_SINGLE,
 *   results = \Drupal\webform\Plugin\WebformHandlerInterface::RESULTS_PROCESSED,
 * )
 */
class WebformHandlerDefer extends WebformHandlerBase {
  const SUBMISSIONS_LOG = 'private://webform_handler_defer_log.json';
  const SUBMISSIONS_LOG_LOCK = 'webform_handler_defer_log';
  const SUBMISSIONS_LOG_LOCK_WAIT = 2.0;

  /**
   * {@inheritdoc}
   *
   * Trigger on submission method 'postSave'.
   * See webform/src/PluginWebformHandlerBase.php 'Submission methods'
   *
   * @todo Consider possible need to intercept events other than 'postSave'.
   */
  public function postSave(WebformSubmissionInterface $submission, $update = TRUE) {
    if (!$update) {
      // Log this submission, and check that was successful.
      $submission_logged = (self::submissionList($submission->id()) !== FALSE);
      // Unless the previous step failed (unlikely), don't run handlers now.
      if ($submission_logged) {
        self::disableHandlers($submission, TRUE);
      }
      else {
        \Drupal::logger('webform_handler_defer')->warning("Submission {$submission->id()} handlers were not deferred.");
      }
    }
  }

  /**
   * Disable relevant handlers.
   */
  private static function disableHandlers(WebformSubmissionInterface $submission, bool $deferred) {
    $webform = $submission->getWebform();
    $this_handler = $webform->getHandler('webform_handler_defer');
    $deferred_handlers = $this_handler->getConfiguration()['settings']['defer_handlers'];

    // For $deferred TRUE, disable the deferred handlers.
    // Otherwise, disable all handlers *except* the deferred handlers.
    if ($deferred) {
      $disable_handlers = $deferred_handlers;
    }
    else {
      $available_handlers = self::getAvailableHandlers($webform);
      $disable_handlers = array_diff_key($available_handlers, $deferred_handlers);
    }

    foreach ($disable_handlers as $handler_id => $handler_label) {
      $handler = $webform->getHandler($handler_id);
      $handler->disable();
      // DEBUG:
      // phpcs:disable
      // \Drupal::messenger()->addMessage('webform_handler_defer: disabled handler: ' . $handler->getLabel());
      // phpcs:enable
    }
  }

  /**
   * Process deferred submit handlers.
   */
  public static function processDeferredHandlers() {
    // Fetch list of previously deferred submissions.
    $submissions = self::submissionList();
    if ($submissions === FALSE) {
      // Failed to fetch deferred submissions list (unlikely).
      \Drupal::logger('webform_handler_defer')->warning("Deferred submission handlers processing will be delayed.");
      return;
    }
    elseif (empty($submissions)) {
      // No deferred submissions - nothing to do.
      return;
    }

    // For proper security and behavior, we run submission handlers using
    // the same account that was used when the form was submitted.
    $account_switcher = \Drupal::service('account_switcher');

    foreach ($submissions as $submission_id) {
      $submission = WebformSubmission::load($submission_id);
      if (empty($submission)) {
        // Handle the case of a submission having been deleted etc.
        continue;
      }

      try {
        // Switch to form submitter's account, or fall back to anonymous user.
        $account_switcher->switchTo($submission->getOwner() ?? User::load(0));
        // Disable all handlers except deferred handlers.
        self::disableHandlers($submission, FALSE);
        // Process deferred handlers.
        $submission->getWebform()->invokeHandlers('postSave', $submission);

        // phpcs:disable
        /*
        // DEBUG:
        $uid = \Drupal::currentUser()->id();
        $msg = '[webform_handler_defer] Processed submission: ' . $submission_id . ' as uid ' . $uid;
        \Drupal::messenger()->addMessage($msg);
        */
        // phpcs:enable
      } finally {
        // Switch back to the original user account.
        $account_switcher->switchBack();
      }
    }
    // Log some basic info about what was done.
    if (count($submissions) > 20) {
      $submissions = array_slice($submissions, 0, 20);
      $submissions[] = 'etc.';
    }
    \Drupal::logger('webform_handler_defer')->info(
      sprintf(
        'Processed deferred handlers for %d submissions [%s]',
        count($submissions),
        implode(', ', $submissions),
      )
    );
  }

  /**
   * Save/fetch deferred submissions IDs.
   *
   * @param int $submission_id
   *   The submission ID to log, or zero/empty to delete the submissions log.
   *
   * @return array|bool
   *   An array of previously deferred submission ids, or FALSE on failure.
   *
   * We use file-based storage for speed, plus Drupal locking for safety.
   * The overall objective: minimize submission time as perceived by the user.
   * @todo Consider alternative methods, focus on write-speed and robustness,
   * e.g. Drupal cache, Queue or TempStore, custom DB table, etc.
   */
  private static function submissionList(int $submission_id = 0): array|bool {
    $file_path = self::SUBMISSIONS_LOG;
    $lock_name = self::SUBMISSIONS_LOG_LOCK;
    $max_wait = self::SUBMISSIONS_LOG_LOCK_WAIT;

    // Avoid simultaneous writes: wait for a successful lock.
    $deadline = microtime(TRUE) + $max_wait;
    while (microtime(TRUE) < $deadline) {
      $lock_acquired = \Drupal::lock()->acquire($lock_name);
      if ($lock_acquired) {
        break;
      }
      // This should happen only rarely, if ever: Sleep 10% of total wait time.
      usleep((1000000 / 10) * $max_wait);
    }

    // In practice, this should "never" happen.
    if (!$lock_acquired) {
      \Drupal::logger('webform_handler_defer')->error('Failed to lock the submissions log.');
      return FALSE;
    }

    try {
      $submission_ids = [];
      // Get previously logged submission ids.
      if (file_exists($file_path)) {
        $json = file_get_contents($file_path);
        $submission_ids = json_decode($json, TRUE);
      }

      // Append the new submission ID if given, otherwise empty the log.
      if (empty($submission_id)) {
        file_put_contents($file_path, json_encode([]));
      }
      else {
        $submission_ids[] = $submission_id;
        file_put_contents($file_path, json_encode($submission_ids));
      }

    } finally {
      // Release the lock.
      \Drupal::lock()->release($lock_name);
    }

    return $submission_ids;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'defer_handlers' => [],
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['defer_handlers'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Handlers to defer'),
      '#description' => $this->t(
        'Selected handlers are processed later, rather than on form submission. Note that some handlers may not be compatible.'
      ),
      '#options' => $this->getAvailableHandlers($this->getWebform()),
      '#default_value' => $this->configuration['defer_handlers'],
    ];
    // Remove our own handler from the options list.
    unset($form['defer_handlers']['#options'][$this->getHandlerId()]);

    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $this->configuration['defer_handlers'] = array_filter($form_state->getValue('defer_handlers'));
    parent::submitConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function getSummary() {
    return [
      '#markup' => $this->t('Deferred: @handlers', [
        '@handlers' => implode(', ', $this->configuration['defer_handlers']),
      ]),
    ];
  }

  /**
   * Helper function to get a list of available handlers on this webform.
   *
   * @return array
   *   An array of handler labels keyed by their plugin IDs.
   */
  private static function getAvailableHandlers(WebformInterface $webform) {
    $options = [];
    foreach ($webform->getHandlers() as $handler) {
      $options[$handler->getHandlerId()] = $handler->label();
    }
    return $options;
  }

}
